% (C) - Boris Bellalta - 2021


function TheMultiPathCase(seed)

rng(seed);

C1 = 10E6;
C2 = 10E6;
L=8000;
KS=100;
B=8E6;

% Sim parameters
TimeHorizon=800;
TimeChange=200;
TimeChange2=400;

N=2; % number of paths

a1=zeros(1,TimeHorizon);
a1(1:TimeChange)=0.1.*ones(1,TimeChange);
a1(TimeChange+1:TimeChange2)=0.3.*ones(1,TimeChange2-TimeChange);
a1(TimeChange2+1:TimeHorizon)=0.1.*ones(1,TimeHorizon-TimeChange2);

b1=zeros(1,TimeHorizon);
b1(1:TimeChange)=0.2.*ones(1,TimeChange);
b1(TimeChange+1:TimeChange2)=0.6.*ones(1,TimeChange2-TimeChange);
b1(TimeChange2+1:TimeHorizon)=0.2.*ones(1,TimeHorizon-TimeChange2);


a2=0.2;
b2=0.4;

w(1,:)=a1 + (b1-a1).*rand(1,TimeHorizon);
w(2,:)=a2 + (b2-a2).*rand(1,TimeHorizon);

% Weights - Actions
%alfa = [0:0.05:1];
%alfa = [0 0.1 0.2 0.3 0.4 0.5 0.6 0.7 0.8 0.9 1];
alfa = [0 0.1 0.25 0.5 0.75 0.9 1];
%alfa = [0.5];

K = length(alfa);

% Gamma
gamma = 0.99;

%----------------- EXP3 ------------------


Q_E3_disp = zeros(K,TimeHorizon);
Q_E3 = zeros(K,1);

R_E3 = zeros(1,TimeHorizon);
alfa_E3 = zeros(1,TimeHorizon);
alfa_selected_E3 = zeros(1,K);

% First Iteration

n_sel = randi(K,1); % Action Selected, random
alfa_E3(1) = n_sel;
alfa_selected_E3(n_sel) = 1;

% First Path
[ED1(1),ELosses1(1)] = PathDelay(KS,alfa(n_sel)*B,L,C1,w(1,1));
% Second Path
[ED2(1),ELosses2(1)] = PathDelay(KS,(1-alfa(n_sel))*B,L,C2,w(1,1));

% Reward (taking only the delay into account)
R_E3(1) = alfa(n_sel)*ED1(1)+(1-alfa(n_sel))*ED2(1); % Mean
%R_EG(1) = max([ED1(1) ED2(1)]); % Max

% Update Q-table
Q_E3(n_sel)=R_E3(1);
Q_E3_Disp(:,1) = Q_E3;

alfa_selected_E3(n_sel) = 1;

Sigma = 0.6; % Controls de Exploration

h=ones(K,TimeHorizon); % weights 

p = zeros(K,TimeHorizon);

for t=2:TimeHorizon
   
    % Calculate distribution
    
    for a=1:K
        p(a,t)=(1-Sigma)*h(a,t-1)/(sum(h(:,t-1))) + Sigma/K;
    end
    
    % Selects an action
    
    P=cumsum(p(:,t));
    aux=rand;
    n_sel = find(aux<P,1,'first'); % find the 1st index s.t. r<D(i);
       
    alfa_E3(t) = n_sel;
    alfa_selected_E3(n_sel)=alfa_selected_E3(n_sel)+1;
    
    % First Path
    [ED1(t),ELosses1(t)] = PathDelay(KS,alfa(n_sel)*B,L,C1,w(1,t));
    % Second Path
    [ED2(t),ELosses2(t)] = PathDelay(KS,(1-alfa(n_sel))*B,L,C2,w(2,t));

    % Reward (taking only the delay into account)
    R_E3(t) = abs(ED1(t)-ED2(t));
        
    Q_E3(n_sel)=(Q_E3(n_sel)*(alfa_selected_E3(n_sel)-1) + R_E3(t))/alfa_selected_E3(n_sel); % Av
    
    Q_E3_disp(:,t)=Q_E3(:); % For display purposes
    
    normQ = (1-(Q_E3(n_sel)/sum(Q_E3)));
    normQ_est = normQ/p(n_sel,t);
    
    % Update weights
    
    for a=1:K
        if(a==n_sel)
            h(a,t)=h(a,t-1)*exp(Sigma*normQ_est/K);
        else
            h(a,t)=h(a,t-1);
        end
    end
    
    h(:,t)=h(:,t)/sum(h(:,t));
    
end

AccR_E3 = cumsum(R_E3); % Note that this time, the minimum is the best

figure
yyaxis left
plot(ED1.*1E3,'k-');
hold
plot(ED2.*1E3,'b-');
hold
ylabel('Delay (ms)');

yyaxis right
plot(alfa_E3,'ro');

ylabel('Selected Action');
xlabel('Time');
legend('Delay Path 1','Delay Path 2','Action Selected');
axis([0 TimeHorizon 0 8])


if(K==7)
figure
plot(h(1,:));
hold
plot(h(2,:));
plot(h(3,:));
plot(h(4,:));
plot(h(5,:));
plot(h(6,:));
plot(h(7,:));
hold
legend('\alpha=0','\alpha=0.1','\alpha=0.25','\alpha=0.5','\alpha=0.75','\alpha=0.9','\alpha=1');
xlabel('Time');
ylabel('Weights h');
end


disp('Final Performance - EXP3');
disp('Max delay (av) between path 1 and path 2 | Total Losses (sum of path 1 and path 2)');
disp([max([mean(ED1) mean(ED2)]) mean(ELosses1)+mean(ELosses2)]);


end




function [ED,ELosses]=PathDelay(K,B,L,C,w)

%disp('Bandwidth | Capacity');
%disp([B C K w]);

if(B > 0)
    ES = L/C;
    lambda=(B/L);
    a = lambda*ES+w;

    if a==1
        EN = K/2;
        Pb = 1/(K+1);
    else
        Pb = (1-a)*a^K/(1-a^(K+1));
        EN = a/(1-a) - ((K+1)*a^(K+1))/(1-a^(K+1));
    end
    ED = EN/(lambda*(1-Pb));
    ELosses = B*Pb;
else
    ED = 0;
    ELosses = 0;    
end
%disp([B ED ELosses]);
%pause

end